/*********************************************************************************
 * Copyright (c) 2008-2010 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/

package eu.unicore.hila.grid.job;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.xmlbeans.XmlCursor;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ApplicationType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.DataStagingType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDefinitionDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDefinitionType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDescriptionType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ResourcesType;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.ArgumentType;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.EnvironmentType;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.POSIXApplicationDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdlPosix.POSIXApplicationType;

import eu.unicore.hila.exceptions.HiLAException;
import eu.unicore.hila.grid.File;

/**
 * A convenience class to help build JSDL documents.
 * 
 * Instantiate an object of this class and use the set and add methods to set or
 * add
 * 
 * <ul>
 * <li>arguments</li>
 * <li>environment variables</li>
 * <li>file exports (after job completion)</li>
 * <li>file imports (before running the job)</li>
 * <li>application name and version</li>
 * <li>executable</li>
 * <li></li>
 * </ul>
 * 
 * Executable and Application (by name and optional version) are mutually
 * exclusive, which is enforced by the methods. Whatever is set last will be
 * used.
 * 
 * @author bjoernh
 * 
 */
public class JSDLBuilder {

    private List<Import> imports;

    private List<Export> exports;

    private String taskName;

    // application xor executable
    private String applicationName;

    private String applicationVersion;

    private String executable;

    private List<String> arguments;

    JobDefinitionDocument jsdl = null;

    private Map<String, String> environment;

    private ResourcesType resources;

    /**
   * 
   */
    public JSDLBuilder() {
	imports = new ArrayList<Import>();
	exports = new ArrayList<Export>();

	arguments = new ArrayList<String>();
	environment = new HashMap<String, String>();

	resources = null;
    }

    /**
     * Get the {@link List} of {@link Import}s already defined.
     * 
     * @return
     */
    public List<Import> getImports() {
	return imports;
    }

    /**
     * Set the {@link List} of {@link Import}s.
     * 
     * @param imports
     */
    public void setImports(List<Import> imports) {
	reset();
	this.imports = imports;
    }

    /**
   * 
   */
    private void reset() {
	jsdl = null;
    }

    /**
     * Get the {@link List} of {@link Export}s already defined.
     * 
     * @return
     */
    public List<Export> getExports() {
	return exports;
    }

    /**
     * Set the list of {@link Export}s.
     * 
     * @param exports
     */
    public void setExports(List<Export> exports) {
	reset();
	this.exports = exports;
    }

    /**
     * Get the name of the application currently defined.
     * 
     * @return The name of the application, if it is defined. <code>null</code>
     *         otherwise.
     */
    public String getApplicationName() {
	return applicationName;
    }

    /**
     * Set the name of the application to be used.
     * 
     * @param applicationName
     */
    public void setApplicationName(String applicationName) {
	reset();
	if (applicationName != null) {
	    executable = null;
	    this.applicationName = applicationName;
	}
    }

    /**
     * Get the version of the application to be used.
     * 
     * @return The version of the application, if defined. <code>null</code>
     *         otherwise.
     */
    public String getApplicationVersion() {
	return applicationVersion;
    }

    /**
     * Set the version of the application to be used.
     * 
     * @param applicationVersion
     */
    public void setApplicationVersion(String applicationVersion) {
	reset();
	this.applicationVersion = applicationVersion;
    }

    public String getExecutable() {
	return executable;
    }

    public void setExecutable(String executable) {
	reset();
	if (executable != null) {
	    applicationName = null;
	    this.executable = executable;
	}
    }

    public List<String> getArguments() {
	return arguments;
    }

    public void setArguments(List<String> arguments) {
	reset();
	this.arguments = arguments;
    }

    public Map<String, String> getEnvironment() {
	return environment;
    }

    public void setEnvironment(Map<String, String> environment) {
	reset();
	this.environment = environment;
    }

    public void setResources(final ResourcesType _resources) {
	reset();
	this.resources = _resources;
    }

    public JobDefinitionDocument getJsdl() throws HiLAException {
	if (jsdl == null) {
	    createJSDL();
	}
	return jsdl;
    }

    // adders
    public void addArgument(String argument) {
	reset();
	arguments.add(argument);
    }

    public void addEnvironment(String variable, String value) {
	reset();
	environment.put(variable, value);
    }

    public void addExport(String fileName, File file) {
	reset();
	exports.add(new Export(fileName, file));
    }

    public void addExport(File file) {
	exports.add(new Export(file.getName(), file));
    }

    public void addImport(File file, String fileName) {
	reset();
	imports.add(new Import(file, fileName));
    }

    public void addImport(File file) {
	addImport(file, file.getName());
    }

    /**
     * @throws HiLAException
     * 
     */
    private void createJSDL() throws HiLAException {
	jsdl = JobDefinitionDocument.Factory.newInstance();

	JobDefinitionType jd = jsdl.addNewJobDefinition();
	JobDescriptionType jDescr = jd.addNewJobDescription();

	POSIXApplicationDocument posixAppD = POSIXApplicationDocument.Factory
		.newInstance();
	POSIXApplicationType posixApp = posixAppD.addNewPOSIXApplication();

	// application or executable
	if ((applicationName == null) ^ (executable == null)) {
	    ApplicationType app = jDescr.addNewApplication();
	    if (applicationName != null) {
		app.setApplicationName(applicationName);
		if (applicationVersion != null) {
		    app.setApplicationVersion(applicationVersion);
		}
	    } else {
		posixApp.addNewExecutable().setStringValue(executable);
	    }

	    // arguments
	    if (arguments != null) {
		for (String argument : arguments) {
		    ArgumentType aType = posixApp.addNewArgument();
		    aType.setStringValue(argument);
		}
	    }

	    // environment
	    if (environment != null) {
		for (Map.Entry<String, String> env : environment.entrySet()) {
		    EnvironmentType eType = posixApp.addNewEnvironment();
		    eType.setName(env.getKey());
		    eType.setStringValue(env.getValue());
		}
	    }

	    XmlCursor cursor = app.newCursor();
	    cursor.toEndToken();
	    posixApp.newCursor().copyXml(cursor);
	} else {
	    throw new HiLAException(
		    "Need exactly one application for execution.");
	}

	if (taskName != null) {
	    jDescr.addNewJobIdentification().setJobName(taskName);
	}

	// data staging
	if (imports != null) {
	    for (Import imprt : imports) {
		DataStagingType dStag = jDescr.addNewDataStaging();
		dStag.addNewSource().setURI(
			imprt.getSource().getLocation().toString());
		dStag.setFileName(imprt.getTargetName());
	    }
	}
	if (exports != null) {
	    for (Export export : exports) {
		DataStagingType dStag = jDescr.addNewDataStaging();
		dStag.addNewTarget().setURI(
			export.getTargetFile().getLocation().toString());
		dStag.setFileName(export.getSourceName());
	    }
	}
	// data staging

	if (resources != null) {
	    jDescr.setResources(resources);
	}
    }

    public String getTaskName() {
	return taskName;
    }

    public void setTaskName(String taskName) {
	reset();
	this.taskName = taskName;
    }

}
