/*********************************************************************************
 * Copyright (c) 2010 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
package eu.unicore.hila.shell.commands;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.concurrent.TimeUnit;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.ParseException;
import org.apache.log4j.Logger;

import eu.unicore.hila.Location;
import eu.unicore.hila.exceptions.HiLAException;
import eu.unicore.hila.grid.Job;
import eu.unicore.hila.grid.File;
import eu.unicore.hila.grid.SimpleTransfer;
import eu.unicore.hila.grid.Site;
import eu.unicore.hila.grid.TaskStatus;
import eu.unicore.hila.grid.job.JSDLBuilder;
import eu.unicore.hila.grid.job.JobDescription;
import eu.unicore.hila.shell.ShellContext;

/**
 * @author bjoernh
 * 
 *         12.05.2010 10:33:01
 * 
 */
public class Run extends ShellCommandBase implements ShellCommand {
    private static final Logger log = Logger.getLogger(Run.class);

    /**
     * 
     */
    public Run() {
	initializeOptions();
    }

    /**
     * 
     */
    private void initializeOptions() {
	options.addOption(OptionConstants.HELP);
	options.addOption(OptionConstants.APP_NAME);
	options.addOption(OptionConstants.APP_VERSION);
	options.addOption(OptionConstants.APP_EXECUTABLE);
	options.addOption(OptionConstants.APP_ARGUMENT);
	options.addOption(OptionConstants.SITE);
    }

    /**
     * @see eu.unicore.hila.shell.commands.ShellCommandBase#execute(java.lang.String[],
     *      eu.unicore.hila.shell.ShellContext)
     */
    @Override
    public void execute(String[] cmdArgs, ShellContext ctx)
	    throws HiLAException {
	InputStream is = null;
	try {
	    CommandLine cl = parser.parse(options, cmdArgs);

	    if (cl.hasOption(OptionConstants.HELP.getArgName())) {
		printHelp(ctx);
		return;
	    }

	    Site executionSite = null;
	    if (cl.hasOption(OptionConstants.SITE.getOpt())) {
		final String siteLocStr = cl
			.getOptionValue(OptionConstants.SITE.getOpt());
		Location siteLoc = new Location(siteLocStr);
		if (!siteLoc.isLocationOfType(Site.class)) {
		    siteLoc = ctx.getCurrentLocation().getChildLocation(
			    siteLocStr);
		}
		if (!siteLoc.isLocationOfType(Site.class)) {
		    log.error("Site location invalid.");
		    return;
		}
		executionSite = (Site) siteLoc.locate();
	    } else {
		log.error("No site given for execution.");
		return;
	    }

	    final JSDLBuilder jb = new JSDLBuilder();

	    // executable
	    if (cl.hasOption(OptionConstants.APP_EXECUTABLE.getOpt())) {
		String aeStr = cl.getOptionValue(OptionConstants.APP_EXECUTABLE
			.getOpt());
		jb.setExecutable(aeStr);
	    }

	    if (cl.hasOption(OptionConstants.APP_ARGUMENT.getOpt())) {
		String[] arguments = cl
			.getOptionValues(OptionConstants.APP_ARGUMENT.getOpt());

		for (String argument : arguments) {
		    jb.addArgument(argument);
		}
	    }

	    JobDescription jd = new JobDescription() {

		@Override
		public Object getActualJobDescription() {
		    try {
			return jb.getJsdl();
		    } catch (HiLAException e) {
			return null;
		    }
		}
	    };

	    Job task = executionSite.submit(jd);
	    task.startSync();

	    while (!isFinalState(task.status())) {
		try {
		    Thread.sleep(1000);
		} catch (InterruptedException e) {

		}
	    }

	    File stdout = task.getStdOut();
	    
	    /*
	    java.io.File tmpStdout = java.io.File.createTempFile("hila", null);

	    SimpleTransfer st = stdout.exportToLocalFile(tmpStdout, true);
	    while (!isFinalState(st.status())) {
		try {
		    Thread.sleep(1000);
		} catch (InterruptedException e) {
		}
	    }

	    is = new FileInputStream(tmpStdout);
	    int b;
	    while ((b = is.read()) != -1) {
		System.out.write(b);
	    }
	    */
	    SimpleTransfer transfer = stdout.exportToStream(System.out);
	    transfer.block();
	} catch (HiLAException e) {

	} catch (ParseException e) {
	    log.error("Unable to parse command line.", e);
	} finally {
	    try {
		if (is != null) {
		    is.close();
		}
	    } catch (IOException e) {
		log.error("Unable to close stream.", e);
	    }
	}
    }

    /**
     * @param status
     * @return
     */
    private boolean isFinalState(TaskStatus status) {
	return status.equals(TaskStatus.ABORTED)
		|| status.equals(TaskStatus.FAILED)
		|| status.equals(TaskStatus.SUCCESSFUL);
    }
}
