/*********************************************************************************
 * Copyright (c) 2009 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
package eu.unicore.hila.grid.unicore6;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.apache.log4j.Logger;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.TargetSystemFactory;
import de.fzj.unicore.uas.client.TSSClient;
import eu.unicore.hila.Location;
import eu.unicore.hila.exceptions.HiLAException;
import eu.unicore.hila.exceptions.HiLALocationSyntaxException;
import eu.unicore.hila.grid.Config;
import eu.unicore.hila.grid.Site;

/**
 * @author bjoernh
 * 
 *         24.08.2009 10:02:06
 * 
 */
public class SiteLocator {
    private static final Logger log = Logger.getLogger(SiteLocator.class);

    private static volatile SiteLocator theInstance = null;
    private static final Object theInstanceLock = new Object();

    private Map<Location, TSSClient> tssClients = new HashMap<Location, TSSClient>();
    private Object tssClientsLock = new Object();

    private Config config = null;
    private Object configLock = new Object();

    public static SiteLocator getInstance() {
	if (theInstance == null) {
	    synchronized (theInstanceLock) {
		if (theInstance == null) {
		    theInstance = new SiteLocator();
		}
	    }
	}
	return theInstance;
    }

    /**
     * 
     * @param registryURLs
     * @param props
     * @param location
     * 
     * @return
     * @throws HiLAException
     */
    @SuppressWarnings("unchecked")
    public List<Site> getAllSites(List<String> registryURLs,
	    Unicore6Properties props, Location location) throws HiLAException {

	/*
	 * timeout can actually become twice as long, because it is used twice.
	 * Minimum timeouts for LAN connections are 2500L if a TargetSystem is
	 * already existing and 4000L if a TargetSystem needs to be created for
	 * remote connections
	 */
	final long timeoutMillis = Long.parseLong(props.getProperty(
		"hila.unicore6.timeout", "5000"));

	if (config == null) {
	    synchronized (configLock) {
		if (config == null) {
		    String configProviderClass = props
			    .getConfigImplementation();
		    Class<Config> configClass;
		    try {
			configClass = (Class<Config>) Class
				.forName(configProviderClass);
		    } catch (ClassNotFoundException e) {
			throw new HiLAException(
				"hila.unicore6.config references unknown class"
					+ configProviderClass, e);
		    } catch (ClassCastException e) {
			throw new HiLAException(configProviderClass
				+ " is not a subclass of "
				+ Config.class.getName());
		    }
		    Constructor<Config> constructor;
		    try {
			constructor = configClass
				.getConstructor(Unicore6Properties.class);
		    } catch (SecurityException e) {
			throw new HiLAException(e.getMessage(), e);
		    } catch (NoSuchMethodException e) {
			throw new HiLAException(
				"Config implementation referenced in "
					+ "hila.unicore6.config does not provide the "
					+ "correct constructor", e);
		    }
		    try {
			config = constructor.newInstance(props);
		    } catch (IllegalArgumentException e) {
			throw new HiLAException(e.getMessage(), e);
		    } catch (InstantiationException e) {
			throw new HiLAException(e.getMessage(), e);
		    } catch (IllegalAccessException e) {
			throw new HiLAException(e.getMessage(), e);
		    } catch (InvocationTargetException e) {
			throw new HiLAException(e.getMessage(), e);
		    }
		}
	    }

	}

	Unicore6SecurityProperties secProps = (Unicore6SecurityProperties) config
		.getIdForLocation(location);

	if (secProps != null) {
	    try {
		secProps = (Unicore6SecurityProperties) secProps.clone();
	    } catch (CloneNotSupportedException e) {
		log.warn("Unable to clone security properties.", e);
	    }
	}

	List<Site> sites = new ArrayList<Site>();
	ExecutorService executor = Executors.newCachedThreadPool();
	List<Future<List<EndpointReferenceType>>> tsfFutures = new ArrayList<Future<List<EndpointReferenceType>>>();
	for (String registryURL : registryURLs) {
	    GetServicesFromRegistry getTSFs = new GetServicesFromRegistry(
		    registryURL, secProps, TargetSystemFactory.TSF_PORT);
	    tsfFutures.add(executor.submit(getTSFs));
	}

	final long tsfDeadline = System.currentTimeMillis() + timeoutMillis;
	// take results and timeout
	for (Future<List<EndpointReferenceType>> tsfFuture : tsfFutures) {
	    try {
		List<Future<List<Site>>> siteFutures = new ArrayList<Future<List<Site>>>();
		List<EndpointReferenceType> tsfEprs = tsfFuture.get(tsfDeadline
			- System.currentTimeMillis(), TimeUnit.MILLISECONDS);
		for (EndpointReferenceType endpointReferenceType : tsfEprs) {
		    GetSitesFromTSF getSites = new GetSitesFromTSF(
			    endpointReferenceType, secProps, location, this);
		    siteFutures.add(executor.submit(getSites));
		}

		final long siteDeadline = System.currentTimeMillis()
			+ timeoutMillis;
		for (Future<List<Site>> siteFuture : siteFutures) {
		    try {
			List<Site> tsfSites = siteFuture.get(siteDeadline
				- System.currentTimeMillis(),
				TimeUnit.MILLISECONDS);
			sites.addAll(tsfSites);
		    } catch (InterruptedException e) {
			log.warn(
				"Interrupted while waiting for list of sites.",
				e);
		    } catch (ExecutionException e) {
			log.warn("Error retrieving target systems.", e);
		    } catch (TimeoutException e) {
			log.warn("Retrieval of target systems timed out.", e);
		    }
		}

	    } catch (InterruptedException e) {
		log.error(
			"Interrupted while waiting for list of TargetSystemFactory.",
			e);
	    } catch (ExecutionException e) {
		log.error("Error retrieving target system factories.", e);
	    } catch (TimeoutException e) {
		log.error("Retrieval of target system factories timed out.", e);
	    }
	}

	executor.shutdown();

	return sites;
    }

    public TSSClient getTSSClient(Location _location) {
	if (tssClients.containsKey(_location)) {
	    return getTSSClientFromCache(_location);
	} else {
	    Location gridLoc;
	    try {
		gridLoc = new Location("unicore6:/");
	    } catch (HiLALocationSyntaxException e1) {
		return null;
	    }
	    try {
		Unicore6Grid grid = (Unicore6Grid) gridLoc.locate();
		if (grid != null) {
		    getAllSites(grid.getProperties().getRegistryUrls(),
			    grid.getProperties(), _location.getParentLocation());
		    return getTSSClientFromCache(_location);
		}
	    } catch (HiLAException e) {
		log.error("");
	    }
	}
	return null;
    }

    private TSSClient getTSSClientFromCache(Location _location) {
	if (tssClients.containsKey(_location)) {
	    synchronized (tssClientsLock) {
		if (tssClients.containsKey(_location)) {
		    return tssClients.get(_location);
		}
	    }
	}
	return null;
    }

    public void putTSSClient(Location _location, TSSClient _client) {
	synchronized (tssClientsLock) {
	    tssClients.put(_location, _client);
	}
    }
}
