/*********************************************************************************
 * Copyright (c) 2009 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/
package eu.unicore.hila.grid.unicore6;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.DataStagingType;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.JobDefinitionDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.ResourcesDocument;
import org.ggf.schemas.jsdl.x2005.x11.jsdl.SourceTargetType;
import org.unigrids.x2006.x04.services.sms.StoragePropertiesDocument;
import org.unigrids.x2006.x04.services.tss.ApplicationResourceType;
import org.unigrids.x2006.x04.services.tss.SubmitDocument;
import org.unigrids.x2006.x04.services.tss.TargetSystemPropertiesDocument;
import org.w3.x2005.x08.addressing.EndpointReferenceType;

import de.fzj.unicore.uas.client.JobClient;
import de.fzj.unicore.uas.client.ReservationClient;
import de.fzj.unicore.uas.client.StorageClient;
import de.fzj.unicore.uas.client.TSSClient;
import de.fzj.unicore.uas.security.IUASSecurityProperties;
import eu.unicore.hila.Location;
import eu.unicore.hila.Metadata;
import eu.unicore.hila.Resource;
import eu.unicore.hila.annotations.ResourceType;
import eu.unicore.hila.common.grid.BaseSite;
import eu.unicore.hila.exceptions.HiLACannotContactSiteException;
import eu.unicore.hila.exceptions.HiLAException;
import eu.unicore.hila.exceptions.HiLALocationSyntaxException;
import eu.unicore.hila.exceptions.HiLANotImplementedException;
import eu.unicore.hila.exceptions.HiLANotSupportedException;
import eu.unicore.hila.exceptions.HiLAResourceAlreadyExistsException;
import eu.unicore.hila.grid.File;
import eu.unicore.hila.grid.Job;
import eu.unicore.hila.grid.Reservation;
import eu.unicore.hila.grid.Site;
import eu.unicore.hila.grid.Storage;
import eu.unicore.hila.grid.job.JobDescription;
import eu.unicore.hila.grid.resources.ResourceDescription;

/**
 * @author bjoernh
 * 
 *         29.05.2009 13:41:35
 * 
 */
@ResourceType(locationStructure = { "unicore6:/sites/{siteName}/?",
	"unicore6:/{user}@sites/{siteName}/?" })
public class Unicore6Site extends BaseSite implements Site {
    private static final Logger log = Logger.getLogger(Unicore6Site.class);

    private static final Map<Location, Unicore6Site> sitesMap = new HashMap<Location, Unicore6Site>();

    private TSSClient tssClient = null;

    private final Metadata metadata;

    private Unicore6ReservationsCollection reservationsCollection;

    private Unicore6Site(Location _location, Object... _extraInformation)
	    throws HiLAException {
	super(_location);
	this.tssClient = SiteLocator.getInstance().getTSSClient(_location);
	this.metadata = new Metadata();
	if (this.tssClient == null) {
	    throw new HiLACannotContactSiteException(
		    "Missing TSS for target system.");
	}
    }

    public static synchronized Unicore6Site locate(Location _location,
	    Object... extraInformation) throws HiLAException {
	if (sitesMap.containsKey(_location)) {
	    return sitesMap.get(_location);
	} else {
	    final Unicore6Site site = new Unicore6Site(_location,
		    extraInformation);
	    sitesMap.put(_location, site);
	    return site;
	}
    }

    /**
     * Add reservations to children, as they're not returned by {@link BaseSite}
     * by default.
     * 
     * @see eu.unicore.hila.common.grid.BaseSite#getChildren()
     * @since 2.2
     */
    @Override
    public synchronized java.util.List<eu.unicore.hila.Resource> getChildren()
	    throws HiLALocationSyntaxException {
	List<Resource> children = super.getChildren();
	if (this.reservationsCollection == null) {
	    this.reservationsCollection = new Unicore6ReservationsCollection(
		    this.getLocation().getChildLocation("reservations"));
	}
	if (!children.contains(this.reservationsCollection)) {
	    children.add(this.reservationsCollection);
	}
	return children;
    };

    /**
     * @see eu.unicore.hila.common.BaseResource#getMetadata()
     */
    @Override
    public Metadata getMetadata() throws HiLAException {
	updateMetadata();

	return this.metadata;
    }

    /**
     * @throws HiLAException
     * @see eu.unicore.hila.grid.Submittable#getTask(java.lang.String)
     */
    public Job getTask(String taskId) throws HiLAException {
	// TODO this is potentially expensive, but serves the purpose for now.
	List<Job> tasks = getTasks();
	for (Job computeTask : tasks) {
	    if (computeTask.getId().equals(taskId)) {
		return computeTask;
	    }
	}
	throw new HiLAException("Task does not exist.");
    }

    /**
     * @throws HiLANotImplementedException
     * @see eu.unicore.hila.grid.Submittable#submit(eu.unicore.hila.grid.job.JobDescription)
     */
    public Job submit(JobDescription jd) throws HiLAException {
	if (jd.getActualJobDescription() instanceof JobDefinitionDocument) {
	    JobDefinitionDocument jdd = (JobDefinitionDocument) jd
		    .getActualJobDescription();

	    DataStagingType[] stagings = jdd.getJobDefinition()
		    .getJobDescription().getDataStagingArray();
	    for (DataStagingType staging : stagings) {
		SourceTargetType src = staging.getSource();
		if (src != null) {
		    Location srcLoc = new Location(src.getURI());
		    if (srcLoc.isLocationOfType(File.class)) {
			try {
			    final Unicore6File f = (Unicore6File) srcLoc
				    .locate();
			    src.setURI("BFT:".concat(f.getStagingURI()));
			} catch (ClassCastException e) {
			    // don't do anything
			}
		    }
		}
		SourceTargetType tgt = staging.getTarget();
		if (tgt != null) {
		    Location tgtLoc = new Location(tgt.getURI());
		    if (tgtLoc.isLocationOfType(File.class)) {
			try {
			    final Unicore6File f = (Unicore6File) tgtLoc
				    .locate();
			    tgt.setURI("BFT:".concat(f.getStagingURI()));
			} catch (ClassCastException e) {
			    // don't do anything
			}
		    }
		}
	    }

	    SubmitDocument sd = SubmitDocument.Factory.newInstance();
	    sd.addNewSubmit().setJobDefinition(jdd.getJobDefinition());

	    try {
		JobClient jobClient = tssClient.submit(sd);
		String taskId = epr2TaskId(jobClient.getEPR());
		Job task = new Unicore6Job(location.getChildLocation("tasks")
			.getChildLocation(taskId), jobClient);
		return task;
	    } catch (Exception e) {
		throw new HiLAException(
			"Error submitting job to target system.", e);
	    }
	} else {
	    throw new HiLANotImplementedException(
		    "UNICORE 6 doesn't support this type of job description.");
	}
    }

    /**
     * @see eu.unicore.hila.common.grid.BaseSite#getStorages()
     */
    @Override
    public List<Storage> getStorages() throws HiLAException {
	return getStorages1();
    }

    /**
     * @see eu.unicore.hila.common.grid.BaseSite#getStorages()
     */
    public List<Storage> getStorages1() throws HiLAException {
	List<Storage> storages = new ArrayList<Storage>();

	List<EndpointReferenceType> storageReferences = tssClient.getStorages();
	for (EndpointReferenceType storageEPR : storageReferences) {
	    try {
		// introduce a res_id -> name map to avoid this extra call for
		// already existing storages
		IUASSecurityProperties storSecProp = this.tssClient
			.getSecurityProperties().clone();
		StorageClient sClient = new StorageClient(storageEPR
			.getAddress().getStringValue(), storageEPR, storSecProp);
		StoragePropertiesDocument spd = sClient
			.getResourcePropertiesDocument();
		String fileSystemName = spd.getStorageProperties()
			.getFileSystem().getName();

		Location storageLocation = this.location.getChildLocation(
			"storages").getChildLocation(fileSystemName);
		Storage storage = (Storage) storageLocation.locate(sClient);

		storages.add(storage);
	    } catch (CloneNotSupportedException e) {
		throw new HiLAException("Unable to clone security properties.",
			e);
	    } catch (Exception e) {
		throw new HiLAException(
			"Unexpected exception while retrieving list of storages.",
			e);
	    }
	}

	return storages;
    }

    /**
     * @throws HiLAException
     * @see eu.unicore.hila.common.grid.BaseSite#getStorage(java.lang.String)
     */
    public Storage getStorage(String storageName) throws HiLAException {
	List<Storage> storages = getStorages();
	for (Storage storage : storages) {
	    if (storage.getName().equals(storageName)) {
		return storage;
	    }
	}
	throw new HiLAException("Storage does not exist.");
    }

    /**
     * @see eu.unicore.hila.grid.Submittable#getTasks()
     */
    public List<Job> getTasks() {
	final List<Job> tasks = new ArrayList<Job>();

	if (tssClient != null) {
	    try {
		List<EndpointReferenceType> jobEprs = tssClient.getJobs();
		for (EndpointReferenceType jobEpr : jobEprs) {
		    String resId = epr2TaskId(jobEpr);
		    try {
			Location taskLocation = location
				.getChildLocation(TasksCollection.TASKS.concat(
					"/").concat(resId));
			Job task = null;
			try {
			    task = new Unicore6Job(taskLocation, new JobClient(
				    jobEpr.getAddress().getStringValue(),
				    jobEpr, tssClient.getSecurityProperties()
					    .clone()));
			} catch (HiLAResourceAlreadyExistsException e) {
			    task = (Job) location.getChildLocation(
				    TasksCollection.TASKS.concat("/").concat(
					    resId)).locate();
			}
			if (log.isDebugEnabled()) {
			    log.debug("Located task ".concat(task.getLocation()
				    .toString()));
			}
			tasks.add(task);
		    } catch (HiLAException e) {
			log.error("Unable to locate task " + resId
				+ " despite being announced by target system.",
				e);
		    } catch (CloneNotSupportedException e) {
			throw new HiLAException(
				"Unable to clone security properties.", e);
		    } catch (Exception e) {
			throw new HiLAException(
				"Unexpected exception while retrieving list of tasks.",
				e);
		    }
		}
	    } catch (Exception e) {
		log.error("Unable to list jobs.", e);
	    }
	}

	return tasks;
    }

    /**
     * TODO determine Task ID from ReferenceParameters
     * 
     * @param jobEpr
     * @return
     */
    public static String epr2TaskId(EndpointReferenceType jobEpr) {
	String resId = jobEpr.getAddress().getStringValue()
		.replaceFirst(".*\\?res=", "");
	if (log.isDebugEnabled()) {
	    log.debug("resId = " + resId);
	}
	return resId;
    }

    /**
     * @throws HiLAException
     * 
     */
    private void updateMetadata() throws HiLAException {
	try {
	    final TargetSystemPropertiesDocument tsp = tssClient
		    .getResourcePropertiesDocument();

	    // applications
	    final List<ApplicationResourceType> applications = new ArrayList<ApplicationResourceType>();
	    for (ApplicationResourceType application : tsp
		    .getTargetSystemProperties().getApplicationResourceArray()) {
		applications.add(application);
	    }
	    metadata.setData("applications", applications);

	    metadata.setData("resourceProperties", tsp);

	} catch (Exception e) {
	    throw new HiLAException(
		    "Unable to retrieve TargetSystemProperties.", e);
	}

    }

    /**
     * @see eu.unicore.hila.grid.Reservable#reserve(ResourceDescription,
     *      Calendar)
     */
    @Override
    public Reservation reserve(ResourceDescription _rd, Calendar _startTime)
	    throws HiLAException {
	// TODO Auto-generated method stub
	try {
	    if (!tssClient.supportsReservation()) {
		throw new HiLANotSupportedException(
			"Server does not support reservation.");
	    }
	} catch (Exception e) {
	    // FIXME this is some awkward code
	    if (e instanceof HiLANotSupportedException) {
		throw (HiLANotSupportedException) e;
	    } else {
		throw new HiLAException(
			"Cannot determine whether server supports reservation.",
			e);
	    }
	}

	ReservationClient resClient = null;
	if (_rd.getActualDescription() instanceof ResourcesDocument) {
	    try {
		resClient = tssClient.createReservationClient(
			(ResourcesDocument) _rd.getActualDescription(),
			_startTime);
	    } catch (Exception e) {
		throw new HiLAException("Unable to create reservation.", e);
	    }
	} else {
	    throw new HiLAException("Unknown resource description.");
	}

	if (resClient != null) {
	    try {
		Location resLoc = getLocation()
			.getChildLocation("reservations").getChildLocation(
				epr2TaskId(resClient.getEPR()));
		Unicore6Reservation u6res = (Unicore6Reservation) resLoc
			.locate(resClient);
		return u6res;
	    } catch (Exception e) {
		throw new HiLAException("Cannot get reservation reference.", e);
	    }

	} else {
	    throw new HiLAException("Unable to create reservation.");
	}
    }

    /**
     * @see eu.unicore.hila.grid.Reservable#getReservations()
     * @since 2.2
     */
    @Override
    public List<Reservation> getReservations() throws HiLAException {
	List<Reservation> reservations = new ArrayList<Reservation>();

	List<EndpointReferenceType> resEPRs = tssClient.getReservations();

	for (EndpointReferenceType resEPR : resEPRs) {
	    ReservationClient resClient;
	    try {
		resClient = new ReservationClient(resEPR.getAddress()
			.getStringValue(), resEPR, tssClient
			.getSecurityProperties().clone());
	    } catch (CloneNotSupportedException e) {
		throw new HiLAException("Something weird with security setup.",
			e);
	    } catch (Exception e) {
		throw new HiLAException("Cannot create reservation client.", e);
	    }
	    Location resLoc = getLocation().getChildLocation("reservations")
		    .getChildLocation(epr2TaskId(resEPR));
	    reservations.add((Reservation) resLoc.locate(resClient));
	}
	return reservations;
    }

    /**
     * @throws HiLAException
     * @see eu.unicore.hila.grid.Reservable#getReservation(java.lang.String)
     * @since 2.2
     */
    @Override
    public Reservation getReservation(String _name) throws HiLAException {
	List<Reservation> reservations = getReservations();
	for (Reservation reservation : reservations) {
	    if (reservation.getName().equals(_name)) {
		return reservation;
	    }
	}
	return null;

    }
}
